import sys
from pathlib import Path

class Importer: 
    extension = ""
    
    def __init__(self, db):
        self.db = db
    
    def do_import(self, params):
        filePath = Path(params)
        
        if filePath.is_dir():
            # only files of extension.
            for currentFile in Path(filePath).glob(self.extension):
                self.import_file(currentFile)
                print("Imported: " + str(currentFile.absolute()))
        elif Path(filePath).is_file():
            self.import_file(filePath)
            print("Imported: " + str(filePath))
        else:
            print("folder or file not found." )
            sys.exit(1)

    def import_file(self, filePath):
        """
        child classes should do work here.
        """
        
    def create_table(self, tableName, columnNames, row1):
        sql = f'DROP TABLE IF EXISTS {tableName}'
        self.db.execute(sql)

        columnCount = len(columnNames)
        columnType = self.get_sqlite_column_type(row1[0])
        sql = f'CREATE TABLE {tableName} ({columnNames[0]} {columnType}'
        
        for i in range(1,columnCount):
            columnType = self.get_sqlite_column_type(row1[i])
            sql += ', ' + columnNames[i] + ' ' + columnType
        sql += ');'

        self.db.execute(sql)

    def insert_table_str(self, tableName, columnNames):
        columnCount = len(columnNames)
        sql = f'INSERT INTO {tableName} ({columnNames[0]}'
        
        for i in range(1,columnCount):
            sql += ', ' + columnNames[i]
        sql += ') VALUES (?'
        for i in range(1,columnCount):
            sql += ', ?'
        sql += ');'

        return(sql)
        
    def get_sqlite_column_type(self, value):
        if isinstance(value, int):
            return 'INT'
        if isinstance(value, float):
            return 'REAL'
        if isinstance(value, str):
            if value.isdigit():
                return 'INT'
            try:
                float(value)
                return 'REAL'
            except ValueError:
                return 'TEXT'
        
        #safe enough default for sqlite
        return 'TEXT'



import openpyxl
import sys
from pathlib import Path
from openpyxl.styles import Alignment, Font

class XlsxExporter: 
    def __init__(self, settings, db):
        self.settings = settings
        self.db = db
        
    def export_tables(self, params):
        sql = """
        SELECT name
        FROM sqlite_master 
        WHERE (type = 'table'
          OR type = 'view')
        AND name like ?
        ORDER BY name;
        """
        
        rows = self.db.query(sql, [params])

        if not rows:
            print("No tables found")
            return

        for row in rows:
            tableName = row['name']
            
            sql = f'SELECT * FROM {tableName};'
            rows = self.db.query(sql)
            
            self.export_rows(self.settings.xlsxDataDump, rows, tableName)
            
            print("Exported " + tableName + " to xlsx data dump file.")

    def delete_all_sheets(self, workbookPath):
        if Path(workbookPath).is_file():
            workbook = openpyxl.load_workbook(workbookPath)
        else:
            workbook = openpyxl.Workbook()    

        for sheetName in workbook.sheetnames:
            workbook.remove(workbook[sheetName])

        #we must have at least 1 sheet to save
        workbook.create_sheet("Sheet")
        
        #save workbook
        try:
            workbook.save(workbookPath)
        except PermissionError:
            print("Uable to save. Permission denied.")    

    def export_rows(self, workbookPath, rows, worksheetName = "Sheet"):
        if Path(workbookPath).is_file():
            workbook = openpyxl.load_workbook(workbookPath)
        else:
            workbook = openpyxl.Workbook()
        
        #remove sheet (delete) if it already exists.
        if worksheetName in workbook.sheetnames:
            workbook.remove(workbook[worksheetName])

        worksheet = workbook.create_sheet(worksheetName)

        if rows:
            #write header
            column = 1
            for key in dict(rows[0]).keys():
                worksheet.cell(column=column, row=1, value=key)
                column = column + 1

            #write rows
            for rowIndex, row in enumerate(rows):
                for colIndex, value in enumerate(row):
                    cell = worksheet.cell(column=colIndex + 1, row=rowIndex + 2)
                    cell.value = row[colIndex]
                    cell.alignment = Alignment(wrapText=True)

        #basic formating
        bold_font = Font(bold=True)

        for cell in worksheet["1:1"]:
            cell.font = bold_font
    
        worksheet.freeze_panes = "A2"
        
        #adjust column widths
        dims = {}
        totalColumnMax = 70 #So long text doesn't make super long columns
        for row in worksheet.rows:
            for cell in row:
                if cell.value:
                    currentColumnMax = dims.get(cell.column_letter, 0)
                    cellLength = len(str(cell.value))
                    dims[cell.column_letter] = min((totalColumnMax, max((currentColumnMax, cellLength))))    
        for col, value in dims.items():
            worksheet.column_dimensions[col].width = value * 1.1 #slightly larger for whatever your font might be.

        #save workbook
        try:
            workbook.save(workbookPath)
        except PermissionError:
            print("Uable to save. Permission denied.")



import sys
from pathlib import Path

from .xlsx_exporter import XlsxExporter

class SqlFileExecuter: 
    outputExtension = ".xlsx"
    
    def __init__(self, settings, db):
        self.settings = settings
        self.db = db 
    
    def do_executions(self, params):
        """
        do_executions has to be the funnest method name I've ever come up with. :-)
        """
    
        if 1 < len(params) and params[1] == "csv":
            self.outputExtension = ".csv"
        
        filePath = Path(params[0])
    
        if filePath.is_dir():
            # only files of sql extension.
            for currentFile in Path(filePath).glob("*.sql"):
                self.execute_sql_file(currentFile)
                print("Executed " + str(currentFile.absolute()))
        elif Path(filePath).is_file():
            self.execute_sql_file(filePath)
            print("Executed: " + str(filePath))
        else:
            print("folder or file not found." )
            sys.exit(1)
            
    def execute_sql_file(self, filePath):
        with open(filePath, newline='') as file:
            allsql = file.read()
            file.close()
        
        statements = allsql.split(";")
        
        needFile = True
        
        for statementIndex, statement in enumerate(statements):
            if not statement:
                continue

            rows = self.db.query(statement)
            
            if not rows:
                continue
            
            if self.outputExtension == ".xlsx":
                if needFile:
                    outputPath = filePath.parent / (filePath.stem + '.xlsx')
                    xlsxExporter = XlsxExporter(self.settings, self.db)
                    xlsxExporter.delete_all_sheets(outputPath)
                    needFile = False;
                
                worksheetName = "Sheet"
                if statementIndex > 0:
                    worksheetName = worksheetName + str(statementIndex)

                xlsxExporter.export_rows(outputPath, rows, worksheetName)
                
            if self.outputExtension == ".csv":
                #TODO
                print("csv needs to be coded")
            
        


import csv
from .importer import Importer
from pathlib import Path

class CsvImporter(Importer): 
    extension = "*.csv"

    def import_file(self, filePath):
        tableName = Path(filePath).stem
        
        with open(filePath, newline='') as file:
            reader = csv.reader(file)
            columnNames = next(reader)
            row1 = next(reader)
            
            self.create_table(tableName, columnNames, row1)
            sql = self.insert_table_str(tableName, columnNames)
            
            self.db.execute(sql, row1)
            
            for row in reader:
                self.db.execute(sql, row)

            self.db.commit()



import openpyxl
from .importer import Importer

class XlsxImporter(Importer): 
    extension = "*.xlsx"
        
    def import_file(self, filePath):
        workbook = openpyxl.load_workbook(filePath, data_only=True)

        for sheetName in workbook.sheetnames:
            # don't import sheets marked to ignore.
            if sheetName.endswith("_ignore"):
                continue
                
            worksheet = workbook.get_sheet_by_name(sheetName)
            lastRow = worksheet.max_row + 1
            lastColumn = worksheet.max_column + 1
    
            columnNames = []
            ignoreColumns = []
            for i in range(1, lastColumn):
                columnName = worksheet.cell(row = 1, column = i).value
                
                if columnName.endswith("_ignore"):
                    ignoreColumns.append(i)
                    continue
                
                columnNames.append(columnName)
                
            sql = self.insert_table_str(sheetName, columnNames)
            
            msgcount = 0
            for r in range(2, lastRow):
                params = []
                for c in range(1, lastColumn):
                    if c in ignoreColumns:
                        continue
                    params.append(worksheet.cell(row = r, column = c).value)
                
                if r == 2:
                    self.create_table(sheetName, columnNames, params)
                    
                self.db.execute(sql, params)    
                
                self.db.commit()
                
                msgcount += 1
                if msgcount > 500:
                    print ('inserted ' + str(r) + ' rows into ' + sheetName)
                    msgcount = 0


import sqlite3
from pathlib import Path

class Database:
    """
    pulled from
    https://stackoverflow.com/questions/38076220/python-mysqldb-connection-in-a-class
    and adapted from there.
    
    This will let you use the Database class either normally like db = Database('db_file.sqlite) or in a with statement:

        with Database('db_file.sqlite') as db:
            # do stuff
            
    and the connection will automatically commit and close when the with statement exits.

    Then, you can encapsulate specific queries that you do often in methods and make them easy to access. For example, if you're dealing with transaction records, you could have a method to get them by date:

        def transactions_by_date(self, date):
            sql = "SELECT * FROM transactions WHERE transaction_date = ?"
            return self.query(sql, (date,))
    
    Here's some sample code where we create a table, add some data, and then read it back out:

        with Database('my_db.sqlite') as db:
            db.execute('CREATE TABLE comments(pkey INTEGER PRIMARY KEY AUTOINCREMENT, username VARCHAR, comment_body VARCHAR, date_posted TIMESTAMP)')
            db.execute('INSERT INTO comments (username, comment_body, date_posted) VALUES (?, ?, current_date)', ('tom', 'this is a comment'))
            comments = db.query('SELECT * FROM comments')
            print(comments)
    """
    def __init__(self, settings):
        self.settings = settings
        
        dbPath = Path(self.settings.databasePath)
        
        self._conn = sqlite3.connect(dbPath)
        self._conn.row_factory = sqlite3.Row
        self._cursor = self._conn.cursor()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def connection(self):
        return self._conn

    @property
    def cursor(self):
        return self._cursor

    def commit(self):
        self.connection.commit()

    def close(self, commit=True):
        if commit:
            self.commit()
        self.connection.close()

    def execute(self, sql, params=None):
        self.cursor.execute(sql, params or ())
        
    def fetchall(self):
        return self.cursor.fetchall()

    def fetchone(self):
        return self.cursor.fetchone()

    def query(self, sql, params=None):
        self.cursor.execute(sql, params or ())
        return self.fetchall()


import openpyxl
import sys
from pathlib import Path
from openpyxl.styles import Alignment, Font

class XlsxDataReplace:
    dataSets = []

    def __init__(self, settings, db):
        self.settings = settings
        self.db = db
    
    def run_script(self, params):
        filePath = Path(params[0])
        filePath = filePath.parent / (filePath.stem + '.xlsx')

        if not filePath.is_file():
            print("File not found: " + str(filePath))
            sys.exit(1)
        
        workbook = openpyxl.load_workbook(filePath)
        
        for dataSet in self.dataSets:
            worksheet = workbook.get_sheet_by_name(dataSet['sheetName'])

            results = self.db.query(dataSet['sql'])

            lastRow = dataSet.get('lastRow', worksheet.max_row)
            
            lastColumn = dataSet['firstColumn'] + len(results[0])           

            for row in range(dataSet['firstRow'], lastRow+1):
                for col in range(dataSet['firstColumn'], lastColumn):
                    cell = worksheet.cell(column=col, row=row)
                    cell.value = None
            
            #write rows
            for rowIndex, resultRow in enumerate(results):
                for colIndex, value in enumerate(resultRow):
                    cell = worksheet.cell(column=dataSet['firstColumn'] + colIndex, row=dataSet['firstRow'] + rowIndex)
                    cell.value = resultRow[colIndex]
            
        try:
            workbook.save(filePath)
        except PermissionError:
            print("Uable to save. Permission denied.")

